from allennlp_models.rc.predictors.bidaf import ReadingComprehensionPredictor
from allennlp_models.rc.predictors.dialog_qa import DialogQAPredictor
from allennlp_models.rc.predictors.transformer_qa import TransformerQAPredictor
